import json
from http.server import HTTPServer, BaseHTTPRequestHandler

from sauna.consumers.base import AsyncConsumer
from sauna.consumers import ConsumerRegister
from sauna import __version__

my_consumer = ConsumerRegister('HTTPServer')


class StoppableHTTPServer(HTTPServer):
    """HTTPServer that stops itself when receiving a threading.Event"""

    def __init__(self, must_stop, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._must_stop = must_stop

    def service_actions(self):
        """Called by the serve_forever() loop.

        Check if Sauna requested the server to shutdown.
        It cannot call self.shutdown() because the server does not run in
        a separated thread.
        """
        if self._must_stop.is_set():
            self._BaseServer__shutdown_request = True


@my_consumer.consumer()
class HTTPServerConsumer(AsyncConsumer):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'port': config.get('port', 8080),
            'address': config.get('address', '')  # listen on all interfaces
        }

    def run(self, must_stop, *args):
        http_server = StoppableHTTPServer(
            must_stop,
            (self.config['address'], self.config['port']),
            Handler
        )
        http_server.serve_forever()
        self.logging('debug', 'Exited consumer thread')

    @staticmethod
    def config_sample():
        return '''
        # HTTP Server that exposes sauna status as a REST API
        - type: HTTPServer
          port: 8080
        '''


class NotFoundError(Exception):
    pass


class Handler(BaseHTTPRequestHandler):

    server_version = 'Sauna/' + __version__

    def do_GET(self):
        data = self.generate_response()
        self.wfile.write(data)

    def do_HEAD(self):
        self.generate_response()

    def generate_response(self):
        try:
            content = self.get_content_from_path()
            code = 200
        except NotFoundError:
            content = {'error': 'Resource not found'}
            code = 404
        data = json.dumps(content).encode()
        self.send_response(code)
        self.send_header('Content-Type', 'application/json')
        self.send_header('Content-Length', len(data))
        self.end_headers()
        return data

    def get_content_from_path(self):
        if self.path == '/':
            status, code = HTTPServerConsumer.get_current_status()
            return {
                'status': status,
                'code': code,
                'checks': HTTPServerConsumer.get_checks_as_dict()
            }
        else:
            raise NotFoundError()

    def log_message(self, format, *args):
        HTTPServerConsumer.logging(
            'debug',
            '{} - {}'.format(self.address_string(), format % args)
        )



from sauna.plugins import Plugin, PluginRegister
from sauna.plugins.ext.http import HTTP
from sauna import DependencyError
import re
import json

my_plugin = PluginRegister('HTTP-JSON')


@my_plugin.plugin()
class HTTPJSON(HTTP):
    def __init__(self, config):
        super().__init__(config)
        try:
            import jsonpath_rw as jsonpath
            self.jsonpath = jsonpath
        except ImportError:
            raise DependencyError(self.__class__.__name__, 'jsonpath_rw',
                                  pypi='jsonpath-rw')

    @my_plugin.check()
    def request(self, check_config):
        code = check_config.get('code', 200)
        expect = check_config.get('expect', None)

        try:
            r = self._do_http_request(check_config)
        except Exception as e:
            return Plugin.STATUS_CRIT, '{}'.format(e)

        if r.status_code != code:
            return (
                Plugin.STATUS_CRIT,
                self._error_message(
                    'Got status code {} instead of {}'.format(
                        r.status_code, code),
                    r, check_config)
            )

        if expect is not None:
            regex = re.compile(expect)
            if 'success_jsonpath' in check_config:
                finder = self.jsonpath.parse(check_config['success_jsonpath'])
                try:
                    data = json.loads(r.text)
                except ValueError as ex:
                    return (
                        Plugin.STATUS_CRIT,
                        self._error_message(
                            'Fail to parse response as JSON: {}'.format(ex),
                            r, check_config)
                    )
                matches = finder.find(data)
                found = any(regex.match(match.value) for match in matches)
            else:
                found = bool(regex.match(r.text))

            if not found:
                return (
                    Plugin.STATUS_CRIT,
                    self._error_message(
                        'Could not find expected result ({})'.format(expect),
                        r, check_config)
                )

        elapsed_ms = int(r.elapsed.microseconds / 1000)
        return (
            self._value_to_status_less(elapsed_ms, check_config),
            'HTTP {} in {} ms'.format(r.status_code, elapsed_ms)
        )

    def _error_message(self, msg, r, check_config):
        error_jsonpath = check_config.get('error_jsonpath', None)
        if error_jsonpath is None:
            return msg
        try:
            data = json.loads(r.text)
        except ValueError:
            return msg
        finder = self.jsonpath.parse(error_jsonpath)
        matches = finder.find(data)
        for m in matches:
            msg += ', {}: {}'.format(m.path, m.value)
        return msg

    @staticmethod
    def config_sample():
        return '''
        # Make an HTTP request and parse the result as JSON
        # timeout, warn and crit are durations in milliseconds
        # If success_jsonpath is undefined, the whole response
        # is matched against the expect regex
        # If error_jsonpath is defined, all matches within the
        # response are returned with the check result
        - type: HTTP-JSON
          checks:
            - type: request
              url: https://www.website.tld/status
              verify_ca_crt: false
              method: GET
              code: 200
              success_jsonpath: '$.status'
              expect: (ok|OK)
              error_jsonpath: '$.message'
              timeout: 5000
              warn: 1000
              crit: 5000
        '''



import time
import fractions
from functools import reduce
import logging


class Scheduler:

    def __init__(self, jobs):
        """
        Create a new Scheduler.

        >>> s = Scheduler([Job(1, max, 100, 200)])
        >>> for jobs in s:
        ...    time.sleep(s.tick_duration)

        :param jobs: Sequence of jobs to schedule
        """
        periodicities = {job.periodicity for job in jobs}
        self.tick_duration = reduce(lambda x, y: fractions.gcd(x, y),
                                    periodicities)
        self._ticks = self.find_minimum_ticks_required(self.tick_duration,
                                                       periodicities)
        self._jobs = jobs
        self._current_tick = 0
        logging.debug('Scheduler has {} ticks, each one is {} seconds'.
                      format(self._ticks, self.tick_duration))

    @staticmethod
    def find_minimum_ticks_required(tick_duration, periodicities):
        """Find the minimum number of ticks required to execute all jobs
        at once."""
        ticks = 1
        for periodicity in reversed(sorted(periodicities)):
            if ticks % periodicity != 0:
                ticks *= int(periodicity / tick_duration)
        return ticks

    def __iter__(self):
        return self

    def __next__(self):
        jobs = [job for job in self._jobs
                if ((self._current_tick * self.tick_duration)
                    % job.periodicity) == 0
                ]
        if jobs:
            logging.debug('Tick {}, scheduled {}'.
                          format(self._current_tick, jobs))
        self._current_tick += 1
        if self._current_tick >= self._ticks:
            self._current_tick = 0
        for job in jobs:
            job()
        return jobs

    def run(self):
        """Shorthand for iterating over all jobs forever.

        >>> print_time = lambda: print(time.time())
        >>> s = Scheduler([Job(1, print_time)])
        >>> s.run()
        1470146095.0748773
        1470146096.076028
        """
        for _ in self:
            time.sleep(self.tick_duration)


class Job:

    def __init__(self, periodicity, func, *func_args, **func_kwargs):
        """
        Create a new Job to be scheduled and run periodically.

        :param periodicity: Number of seconds to wait between job runs
        :param func: callable that perform the job action
        :param func_args: arguments of the callable
        :param func_kwargs: keyword arguments of the callable
        """
        if not callable(func):
            raise ValueError('func attribute must be callable')
        self.periodicity = periodicity
        self.func = func
        self.func_args = func_args
        self.func_kwargs = func_kwargs

    def __repr__(self):
        try:
            name = self.func.__name__
        except AttributeError:
            name = 'unknown'
        return '<Job {} every {} seconds>'.format(name,
                                                  self.periodicity)

    def __call__(self, *args, **kwargs):
        self.func(*self.func_args, **self.func_kwargs)



from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('Dummy')


@my_plugin.plugin()
class Dummy(Plugin):

    @my_plugin.check()
    def dummy(self, check_config):
        return (check_config.get('status', 0),
                check_config.get('output', 'OK'))

    @staticmethod
    def config_sample():
        return '''
        # Fake checks that return the provided
        # status and output
        - type: Dummy
          checks:
            - type: dummy
              status: 0
              output: Everything is alright
        '''



from sauna.commands import CommandRegister

list_commands = CommandRegister()


@list_commands.command(name='list-active-checks')
def list_active_checks(sauna_instance, args):
    """Display the checks that sauna will run."""
    for name in sorted(sauna_instance.get_active_checks_name()):
        print(name)


@list_commands.command(name='list-available-checks')
def list_available_checks(sauna_instance, args):
    """Display the available checks."""
    for plugin, checks in sorted(
            sauna_instance.get_all_available_checks().items()
    ):
        print('{}: {}'.format(plugin, ', '.join(checks)))


@list_commands.command(name='list-available-consumers')
def list_available_consumers(sauna_instance, args):
    """Display the available consumers."""
    for c in sorted(sauna_instance.get_all_available_consumers()):
        print(c)



from collections import namedtuple
import threading
import queue
import logging
import time
import socket
import os
import textwrap
import signal
import importlib
import pkgutil
import re
import sys
import glob
import functools

from sauna import plugins, consumers
from sauna.plugins.base import Check
from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister
from sauna.plugins import PluginRegister
from sauna.scheduler import Scheduler, Job

__version__ = '0.0.16'

ServiceCheck = namedtuple('ServiceCheck',
                          ['timestamp', 'hostname', 'name',
                           'status', 'output'])

# Global dict containing the last status of each check
# Needs to hold a lock to access it
check_results = {}
check_results_lock = threading.Lock()

try:
    # In Python 3.2 threading.Event is a factory function
    # the real class lives in threading._Event
    event_type = threading._Event
except AttributeError:
    event_type = threading.Event


class DependencyError(Exception):

    def __init__(self, plugin, dep_name, pypi='', deb=''):
        self.msg = '{} depends on {}. It can be installed with:\n'.format(
            plugin, dep_name
        )
        if pypi:
            self.msg = '{}    pip install {}\n'.format(self.msg, pypi)
        if deb:
            self.msg = '{}    apt-get install {}\n'.format(self.msg, deb)

    def __str__(self):
        return self.msg


def _merge_config(original, included):
    """Add properties from a dict to another dict.

    :param original: dict to update
    """
    for key, value in included.items():
        if isinstance(value, list):
            try:
                original[key].extend(value)
            except KeyError:
                original[key] = value
        elif isinstance(value, dict):
            try:
                original[key].update(value)
            except KeyError:
                original[key] = value
        else:
            original[key] = value


def read_config(config_file):
    # importing yaml here because dependency is not installed
    # when fetching __version__ from setup.py
    import yaml

    try:
        with open(config_file) as f:
            config = yaml.safe_load(f)
    except OSError as e:
        print('Cannot read configuration file {}: {}'
              .format(config_file, e))
        exit(1)

    for config_file_included in glob.glob(config.get('include', '')):
        config_included = read_config(config_file_included)
        _merge_config(config, config_included)

    return config


class Sauna:

    def __init__(self, config=None):
        if config is None:
            config = {}
        self.config = config
        self.must_stop = threading.Event()
        self._consumers_queues = []
        self.import_submodules(__name__ + '.plugins.ext')
        self.import_submodules(__name__ + '.consumers.ext')
        for extra_plugin_path in self.config.get('extra_plugins', []):
            self.import_directory_modules(extra_plugin_path)

    @classmethod
    def assemble_config_sample(cls, path):
        sample = '---\nperiodicity: 120\nhostname: node-1.domain.tld\n'

        sample += '\nconsumers:\n'
        consumers_sample = ''
        for _, consumer_info in ConsumerRegister.all_consumers.items():
            if hasattr(consumer_info['consumer_cls'], 'config_sample'):
                consumers_sample += textwrap.dedent(
                    consumer_info['consumer_cls'].config_sample()
                )
        sample += consumers_sample.replace('\n', '\n  ')

        sample += '\nplugins:\n'
        plugins_sample = ''
        for _, plugin_info in PluginRegister.all_plugins.items():
            if hasattr(plugin_info['plugin_cls'], 'config_sample'):
                plugins_sample += textwrap.dedent(
                    plugin_info['plugin_cls'].config_sample()
                )
        sample += plugins_sample.replace('\n', '\n  ')

        file_path = os.path.join(path, 'sauna-sample.yml')
        with open(file_path, 'w') as f:
            f.write(sample)

        return file_path

    @property
    @functools.lru_cache()
    def hostname(self):
        # socket.getfqdn can be a very long call
        # make sure to only call it when absolutely necessary
        # that's why a cache is used and dict.get() is avoided
        try:
            return self.config['hostname']
        except KeyError:
            return socket.getfqdn()

    @property
    def periodicity(self):
        return self.config.get('periodicity', 120)

    @property
    def plugins_checks(self):
        plugins = []
        if type(self.config['plugins']) is dict:
            for plugin_name, plugin_data in self.config['plugins'].items():
                plugin_data.update({'type': plugin_name})
                plugins.append(plugin_data)
        elif type(self.config['plugins']) is list:
            plugins = self.config['plugins']
        else:
            print('Invalid configuration, plugins must be a list or a dict')
            exit(1)
        return plugins

    @property
    def consumers(self):
        consumers = []
        if type(self.config['consumers']) is dict:
            for cons_name, cons_data in self.config['consumers'].items():
                # Consumer Stdout doesn't need configuration
                if cons_data is None:
                    cons_data = {}
                cons_data.update({'type': cons_name})
                consumers.append(cons_data)
        elif type(self.config['consumers']) is list:
            consumers = self.config['consumers']
        else:
            print('Invalid configuration, consumers must be a list or a dict')
            exit(1)
        return consumers

    def get_active_checks_name(self):
        checks = self.get_all_active_checks()
        return [check.name for check in checks]

    def get_all_available_consumers(self):
        return_consumers = []
        for plugin_name, _ in ConsumerRegister.all_consumers.items():
            return_consumers.append(plugin_name)

        return return_consumers

    def get_all_available_checks(self):
        checks = {}
        for plugin_name, data in PluginRegister.all_plugins.items():
            checks[plugin_name] = []
            for check in data['checks']:
                checks[plugin_name].append(check)
        return checks

    def get_all_active_checks(self):
        checks = []
        deps_error = []
        for plugin_data in self.plugins_checks:
            plugin_name = plugin_data['type']
            # Load plugin
            plugin_info = PluginRegister.get_plugin(plugin_name)
            if not plugin_info:
                print('Plugin {} does not exist'.format(plugin_name))
                exit(1)

            # Configure plugin
            try:
                plugin = plugin_info['plugin_cls'](
                    plugin_data.get('config', {})
                )
            except DependencyError as e:
                deps_error.append(str(e))
                continue

            # Launch plugin checks
            for check in plugin_data['checks']:
                func_name = plugin_info['checks'].get(check['type'])
                if func_name is None:
                    print('Unknown check {} on plugin {}'.format(check['type'],
                                                                 plugin_name))
                    exit(1)
                check_func = getattr(plugin, func_name)

                # An empty string is a valid check name
                check_name = check.get(
                    'name',
                    '{}_{}'.format(plugin_name, check['type'])
                ).lower()

                check_periodicity = (check.get('periodicity') or
                                     self.periodicity)

                checks.append(Check(check_name, check_periodicity,
                                    check_func, check))
        if deps_error:
            for error in deps_error:
                print(error)
            exit(1)

        # Check duplicate name
        names = [check.name for check in checks]
        duplicates_names = {name: names.count(name)
                            for name in names if names.count(name) > 1}
        for name, count in duplicates_names.items():
            print("check name {} was found {} times, please add name"
                  " field to theses checks".format(name, count))
        if duplicates_names:
            exit(1)
        return checks

    def launch_all_checks(self):
        """Run once every single check."""
        for check in self.get_all_active_checks():
            yield self.launch_check(check)

    def launch_and_publish_checks_with_periodicity(self, periodicity):
        """Run once every check matching the given periodicity.

        Sends the result of each check to the queues and shared dict.
        """
        checks = filter(lambda x: x.periodicity == periodicity,
                        self.get_all_active_checks())
        for check in checks:
            service_check = self.launch_check(check)
            logging.debug('Pushing to consumers: {}'.
                          format(service_check.name))
            self.send_data_to_consumers(service_check)
            with check_results_lock:
                check_results[service_check.name] = service_check

    def launch_check(self, check):
        try:
            status, output = check.run_check()
        except Exception as e:
            logging.warning('Could not run check {}: {}'.format(
                check.name, str(e)
            ))
            status = 3
            output = str(e)
        return ServiceCheck(
            timestamp=int(time.time()),
            hostname=self.hostname,
            name=check.name,
            status=status,
            output=output
        )

    def run_producer(self):
        periodicities = {check.periodicity
                         for check in self.get_all_active_checks()}
        jobs = [Job(p, self.launch_and_publish_checks_with_periodicity, p)
                for p in periodicities]
        logging.info('Running checks with interval: {}'
                     .format(str(periodicities)))
        scheduler = Scheduler(jobs)

        for _ in scheduler:
            if self.must_stop.wait(timeout=scheduler.tick_duration):
                break
        logging.debug('Exited producer thread')

    def term_handler(self, *args):
        """Notify producer and consumer that they should stop."""
        if not self.must_stop.is_set():
            self.must_stop.set()
            self.send_data_to_consumers(self.must_stop)
            logging.info('Exiting...')

    def send_data_to_consumers(self, data):
        for queue in self._consumers_queues:
            queue.put(data)

    def launch(self):
        # Start producer and consumer threads
        producer = threading.Thread(
            name='producer', target=self.run_producer
        )
        producer.start()

        consumers_threads = []
        for consumer_data in self.consumers:

            consumer_name = consumer_data['type']

            consumer_info = ConsumerRegister.get_consumer(consumer_name)
            if not consumer_info:
                print('Plugin {} does not exist'.format(consumer_name))
                exit(1)

            try:
                consumer = consumer_info['consumer_cls'](consumer_data)
            except DependencyError as e:
                print(str(e))
                exit(1)

            if isinstance(consumer, QueuedConsumer):
                consumer_queue = queue.Queue()
                self._consumers_queues.append(consumer_queue)
            else:
                consumer_queue = None

            consumer_thread = threading.Thread(
                name='consumer_{}'.format(consumer_name),
                target=consumer.run,
                args=(self.must_stop, consumer_queue)
            )

            consumer_thread.start()
            consumers_threads.append(consumer_thread)
            logging.debug(
                'Running consumer {}'.format(consumer_name)
            )

        signal.signal(signal.SIGTERM, self.term_handler)
        signal.signal(signal.SIGINT, self.term_handler)

        producer.join()
        self.term_handler()

        for consumer_thread in consumers_threads:
            consumer_thread.join()

        logging.debug('Exited main thread')

    @classmethod
    def import_submodules(cls, entity):
        """Import packages and/or modules."""
        entity = importlib.import_module(entity)
        try:
            for _, name, is_pkg in pkgutil.walk_packages(entity.__path__):
                if not name.startswith('_') and is_pkg is False:
                    full_name = entity.__name__ + '.' + name
                    importlib.import_module(full_name)
        except AttributeError:
            pass

    @classmethod
    def import_directory_modules(cls, path):
        """Import all modules from a filesystem directory."""
        try:
            entries = os.listdir(path)
        except OSError as e:
            logging.error('Cannot load plugins from {}: {}'.format(path, e))
            return
        logging.info('Loading extra plugins from {}'.format(path))
        sys.path.append(path)
        for entry in entries:
            if not os.path.isfile(os.path.join(path, entry)):
                continue
            regexp_result = re.search(r'(.+)\.py$', entry)
            if regexp_result:
                logging.debug('Loading {}'.format(regexp_result.groups()[0]))
                cls.import_submodules(regexp_result.groups()[0])



import socket

from sauna.plugins import (Plugin, PluginRegister)

my_plugin = PluginRegister('TCP')


@my_plugin.plugin()
class Tcp(Plugin):

    @my_plugin.check()
    def request(self, check_config):
        try:
            with socket.create_connection((check_config['host'],
                                           check_config['port']),
                                          timeout=check_config['timeout']):
                pass
        except Exception as e:
            return Plugin.STATUS_CRIT, "{}".format(e)
        else:
            return Plugin.STATUS_OK, "OK"

    @staticmethod
    def config_sample():
        return '''
        # Tcp
        - type: TCP
          checks:
            - type: request
              host: localhost
              port: 11211
              timeout: 5
        '''



import re

from sauna.plugins import PluginRegister
from sauna.plugins.base import PsutilPlugin

my_plugin = PluginRegister('Processes')


@my_plugin.plugin()
class Processes(PsutilPlugin):

    @my_plugin.check()
    def count(self, check_config):
        num_pids = len(self.psutil.pids())
        return (
            self._value_to_status_less(num_pids, check_config),
            '{} processes'.format(num_pids)
        )

    @my_plugin.check()
    def zombies(self, check_config):
        zombies = [p for p in self.psutil.process_iter()
                   if p.status() == 'zombie']
        num_zombies = len(zombies)
        return (
            self._value_to_status_less(num_zombies, check_config),
            '{} zombies'.format(num_zombies)
        )

    def _count_running_processes(self, check_config):
        """Count the number of times a process is running.

        Processes are identified by their first argument 'exec' and
        additional 'args'.
        :rtype int
        """
        process_exec = check_config['exec']
        required_args = check_config.get('args', '').split()
        instances = 0

        for process in self.psutil.process_iter():
            try:
                cmdline = process.cmdline()
            except (self.psutil.NoSuchProcess, self.psutil.AccessDenied):
                # Zombies and processes that stopped throw NoSuchProcess
                continue

            # Often cmdline in an empty list
            if not cmdline:
                continue

            if cmdline[0] != process_exec:
                continue

            if self._required_args_are_in_cmdline(required_args, cmdline):
                instances += 1

        return instances

    @my_plugin.check()
    def running(self, check_config):
        process = check_config['exec']
        nb = check_config.get('nb')
        instances = self._count_running_processes(check_config)

        if instances == 0:
            return self.STATUS_CRIT, 'Process {} not running'.format(process)
        if nb is None:
            return self.STATUS_OK, 'Process {} is running'.format(process)
        if instances == nb:
            return self.STATUS_OK, 'Process {} is running'.format(process)
        return (self.STATUS_WARN,
                'Process {} is running {} times'.format(process, instances))

    @my_plugin.check()
    def file_descriptors(self, check_config):
        """Check that processes and system are not running out of fd."""
        check_config = self._strip_percent_sign_from_check_config(check_config)

        # Check fds of individuals processes
        names, worst_value = self._get_processes_exhausting_fds(check_config)
        if names:
            return (
                self._value_to_status_less(worst_value, check_config),
                'Processes running out of fd: {}'.format(', '.join(names))
            )

        # Check fds of whole system
        percent_used_fds = self._get_percent_system_used_fds()
        if (self._value_to_status_less(percent_used_fds, check_config) !=
                self.STATUS_OK):
            return (
                self._value_to_status_less(percent_used_fds, check_config),
                'System using {}% of file descriptors'.format(percent_used_fds)
            )

        return self.STATUS_OK, 'File descriptors under the limits'

    def _get_processes_exhausting_fds(self, check_config):
        processes_names = set()
        highest_percentage = 0
        for process in self.psutil.process_iter():
            try:
                open_fd = process.num_fds()
                limit_fd = self._get_process_fd_limit(process.pid)
                percentage = int(open_fd * 100 / limit_fd)
                if (self._value_to_status_less(percentage, check_config) !=
                        self.STATUS_OK):
                    processes_names.add(process.name())
                if percentage > highest_percentage:
                    highest_percentage = percentage
            except (self.psutil.NoSuchProcess, self.psutil.AccessDenied,
                    OSError):
                pass
        return processes_names, highest_percentage

    @classmethod
    def _get_process_fd_limit(cls, pid):
        """Retrieve the soft limit of usable fds for a process."""
        with open('/proc/{}/limits'.format(pid)) as f:
            limits = f.read()
        match = re.search(r'^Max open files\s+(\d+)\s+\d+\s+files',
                          limits, flags=re.MULTILINE)
        if not match:
            raise Exception('Cannot parse /proc/{}/limits'.format(pid))
        return int(match.group(1))

    @classmethod
    def _get_percent_system_used_fds(cls):
        """Percentage of opened file descriptors over the whole system.

        :rtype int
        """
        with open('/proc/sys/fs/file-nr') as f:
            # Weird behavior on Python 3.2 on Wheezy:
            # f.read() does not return the entire content of this
            # particular file while it should according to the docs.
            # As the file only contains a single line, f.readline()
            # does the job.
            file_nr = f.readline()
        match = re.match(r'^(\d+)\t\d+\t(\d+)$', file_nr)
        if not match:
            raise Exception('Cannot parse /proc/sys/fs/file-nr')
        system_opened_fds = int(match.group(1))
        system_max_fds = int(match.group(2))
        return int(system_opened_fds * 100 / system_max_fds)

    @classmethod
    def _required_args_are_in_cmdline(cls, required_args, cmdline):
        for arg in required_args:
            if arg not in cmdline[1:]:
                return False
        return True

    @staticmethod
    def config_sample():
        return '''
        # Information about processes
        - type: Processes
          checks:
            # Number of processes in the system
            - type: count
              warn: 400
              crit: 500
            # Number of zombies processes in the system
            - type: zombies
              warn: 1
              crit: 5
            # File descriptors
            - type: file_descriptors
              warn: 60%
              crit: 80%
            # Critical if process is not running
            - type: running
              name: docker_running
              exec: /usr/bin/docker
              args: daemon
        '''



import xmlrpc.client
import http.client
import socket

from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('Supervisor')


@my_plugin.plugin()
class Supervisor(Plugin):

    def __init__(self, config):
        super().__init__(config)

        serverurl = config.get('serverurl', 'unix:///var/run/supervisor.sock')
        timeout = config.get('timeout', 5)
        if serverurl.startswith('unix://'):
            serverurl = serverurl.replace('unix://', '', 1)
            # xmlrpc.client does not support Unix sockets, so we must provide
            # a custom transport layer
            transport = UnixStreamTransport(serverurl, timeout=timeout)
            server = xmlrpc.client.ServerProxy('http://noop',
                                               transport=transport)
        else:
            transport = CustomHTTPTransport(timeout=timeout)
            server = xmlrpc.client.ServerProxy(serverurl, transport=transport)

        rpc_namespace = config.get('rpc_namespace', 'supervisor')
        self.supervisor = getattr(server, rpc_namespace)
        self.supervisor_addr = serverurl

    @my_plugin.check()
    def service(self, check_config):
        try:
            service = check_config['service']
        except KeyError:
            raise KeyError('A name parameter is required for Supervisor '
                           'service checks')
        states_threshold = self._get_states_threshold(check_config)

        try:
            service_info = self.supervisor.getProcessInfo(service)
        except Exception as ex:
            raise Exception('Error while contacting Supervisor at {}: {}'
                            .format(self.supervisor_addr, ex))

        service_state = service_info['statename']
        status = self._get_status(service_state, states_threshold)

        return status, 'Service {} {}'.format(service, service_state)

    @my_plugin.check()
    def services(self, check_config):
        states_threshold = self._get_states_threshold(check_config)
        whitelist = check_config.get('whitelist', [])
        blacklist = check_config.get('blacklist', [])

        def service_enabled(name):
            if blacklist and name in blacklist:
                return False
            if whitelist and name not in whitelist:
                return False
            return True

        try:
            service_infos = self.supervisor.getAllProcessInfo()
        except Exception as ex:
            raise Exception('Error while contacting Supervisor at {}: {}'
                            .format(self.supervisor_addr, ex))

        service_states = {info['name']: info['statename']
                          for info in service_infos
                          if service_enabled(info['name'])}
        service_statuses = {name: self._get_status(state, states_threshold)
                            for name, state in service_states.items()}

        status_priority = [Plugin.STATUS_OK, Plugin.STATUS_UNKNOWN,
                           Plugin.STATUS_WARN, Plugin.STATUS_CRIT]
        status_priority = {s: i for i, s in enumerate(status_priority)}

        global_status = Plugin.STATUS_OK
        for status in service_statuses.values():
            if status_priority[status] > status_priority[global_status]:
                global_status = status

        faulty_services = [name for name, status in service_statuses.items()
                           if status != Plugin.STATUS_OK]
        if faulty_services:
            msg = 'Found {} services out of {} with incorrect state: '\
                  .format(len(faulty_services), len(service_statuses)) +\
                  ', '.join('{} is {}'.format(name, service_states[name])
                            for name in faulty_services)
        else:
            msg = 'All {} services OK'.format(len(service_statuses))

        return global_status, msg

    @staticmethod
    def _get_status(state, states_threshold):
        status_name = states_threshold.get(state, 'UNKNOWN')
        status = getattr(Plugin, 'STATUS_' + status_name,
                         Plugin.STATUS_UNKNOWN)
        return status

    @staticmethod
    def _get_states_threshold(check_config):
        user_threshold = check_config.get('states', {})
        states_threshold = {
            'STARTING': 'OK',
            'RUNNING': 'OK',
            'BACKOFF': 'WARN',
            'STOPPING': 'WARN',
            'STOPPED': 'CRIT',
            'FATAL': 'CRIT'
            # Other states are translated to unknown
        }
        states_threshold.update({a.upper(): b.upper()
                                 for a, b in user_threshold.items()})
        return states_threshold

    @staticmethod
    def config_sample():
        return '''
        # Monitor services of a Supervisor instance
        # Each check can monitor one Supervisor process (type 'service') or all
        # of them (type 'services')
        # The serverurl option should be the same as the one from the
        # Supervisor config, with optional credentials added for HTTP(S)
        - type: Supervisor    # Minimal config to check local services
          checks:
            - type: services
        - type: Supervisor
          config:
            serverurl: unix:///var/run/supervisor.sock  # Same as default
          checks:
            - type: service
              name: supervisor_foo
              service: foo
              states:
                STOPPED: OK  # Allowed: UNKOWN, OK, WARN and CRIT (default)
            - type: services
              name: supervisor_services
              blacklist: [foo]
        - type: Supervisor
          config:
            serverurl: http://bob:pa$$word@example.com:9001/RPC2
            rpc_namespace: ubervisor    # Default is supervisor
            timeout: 10                 # Default is 5s
          checks:
            - type: services
              name: ubervisor_services
              whitelist: [bar,spam]
        '''


# Thanks to Adaptation (http://stackoverflow.com/a/23837147)
class UnixStreamHTTPConnection(http.client.HTTPConnection):
    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(self.host)
        self.sock = sock


class UnixStreamTransport(xmlrpc.client.Transport):
    def __init__(self, socket_path, **conn_kwargs):
        self.socket_path = socket_path
        self._conn_kwargs = conn_kwargs
        super(UnixStreamTransport, self).__init__()

    def make_connection(self, host):
        return UnixStreamHTTPConnection(self.socket_path, **self._conn_kwargs)


class CustomHTTPTransport(xmlrpc.client.Transport):
    def __init__(self, **conn_kwargs):
        self._conn_kwargs = conn_kwargs
        super(CustomHTTPTransport, self).__init__()

    def make_connection(self, host):
        # return an existing connection if possible. This allows
        # HTTP/1.1 keep-alive.
        if self._connection and host == self._connection[0]:
            return self._connection[1]
        # create a HTTP connection object from a host descriptor
        chost, self._extra_headers, x509 = self.get_host_info(host)
        self._connection = (
            host,
            http.client.HTTPConnection(chost, **self._conn_kwargs)
        )
        return self._connection[1]



from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('HTTP')


@my_plugin.plugin()
class HTTP(Plugin):

    def __init__(self, config):
        super().__init__(config)
        try:
            import requests
            self.requests = requests
        except ImportError:
            from ... import DependencyError
            raise DependencyError(self.__class__.__name__, 'requests',
                                  'requests', 'python3-requests')

    @my_plugin.check()
    def request(self, check_config):
        code = check_config.get('code', 200)
        content = check_config.get('content', '')

        try:
            r = self._do_http_request(check_config)
        except Exception as e:
            return Plugin.STATUS_CRIT, '{}'.format(e)

        if r.status_code != code:
            return (
                Plugin.STATUS_CRIT,
                'Got status code {} instead of {}'.format(r.status_code, code)
            )
        if content not in r.text:
            return (
                Plugin.STATUS_CRIT,
                'Content "{}" not in response'.format(content)
            )
        elapsed_ms = int(r.elapsed.microseconds / 1000)
        return (
            self._value_to_status_less(elapsed_ms, check_config),
            'HTTP {} in {} ms'.format(r.status_code, elapsed_ms)
        )

    def _do_http_request(self, check_config):
        method = check_config.get('method', 'GET').upper()
        timeout = check_config.get('timeout', 10000) / 1000
        verify_ca_crt = check_config.get('verify_ca_crt', True)
        url = check_config['url']

        return self.requests.request(method, url,
                                     verify=verify_ca_crt,
                                     timeout=timeout)

    @staticmethod
    def config_sample():
        return '''
        # Make an HTTP request
        # timeout, warn and crit are durations in milliseconds
        - type: HTTP
          checks:
            - type: request
              url: https://www.website.tld
              verify_ca_crt: true
              method: GET
              code: 200
              content: Welcome!
              timeout: 5000
              warn: 1000
              crit: 5000
        '''



from sauna.commands import CommandRegister

status_commands = CommandRegister()


@status_commands.command(name='status')
def list_active_checks(sauna_instance, args):
    """Show the result of active checks."""
    human_status = {
        0: 'OK',
        1: 'Warning',
        2: 'Critical',
        3: 'Unknown',
    }
    for check in sorted(sauna_instance.launch_all_checks()):
        print('  {:<30} {:^14} {}'.format(
            check.name, human_status[check.status], check.output
        ))



#!/usr/bin/env python3
"""Daemon that runs and reports health checks

Usage:
  sauna [--level=<lvl>] [--config=FILE] [<command> <args>...]
  sauna sample
  sauna (-h | --help)
  sauna --version

Options:
  -h --help      Show this screen.
  --version      Show version.
  --level=<lvl>  Log level [default: warn].
  --config=FILE  Config file [default: sauna.yml].

Available commands:
"""
import logging

from docopt import docopt, DocoptLanguageError
from yaml.error import YAMLError

import sauna
from sauna import commands


def build_main_doc():
    sauna.Sauna.import_submodules('sauna.commands.ext')
    doc = __doc__
    for name, func in sorted(commands.CommandRegister.all_commands.items()):
        summary = func.__doc__ .splitlines()[0]
        doc += '  {:<28}  {}\n'.format(name, summary)
    return doc


def main():
    doc = build_main_doc()
    args = docopt(doc, version=sauna.__version__, options_first=True)
    conf_file = args['--config']
    logging.basicConfig(
        format='%(asctime)s - %(levelname)-8s - %(message)s',
        datefmt='%Y/%m/%d %H:%M:%S',
        level=getattr(logging, args['--level'].upper(), 'WARN')
    )
    logging.getLogger('requests').setLevel(logging.ERROR)

    # Sample command needs a not configured instance of sauna
    if args.get('<command>') == 'sample':
        sauna_instance = sauna.Sauna()
        file_path = sauna_instance.assemble_config_sample('./')
        print('Created file {}'.format(file_path))
        exit(0)

    try:
        config = sauna.read_config(conf_file)
    except YAMLError as e:
        print('YAML syntax in configuration file {} is not valid: {}'.
              format(conf_file, e))
        exit(1)
    sauna_instance = sauna.Sauna(config)

    # Generic commands implemented in sauna.commands package
    if args.get('<command>'):
        argv = [args['<command>']] + args['<args>']
        try:
            func = commands.CommandRegister.all_commands[args['<command>']]
        except KeyError:
            print('{} is not a valid command'.format(args['<command>']))
            exit(1)
        try:
            command_args = docopt(func.__doc__, argv=argv)
        except DocoptLanguageError:
            command_args = None
        func(sauna_instance, command_args)

    # Just run sauna
    else:
        sauna_instance.launch()



import time
from datetime import timedelta
import os

from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('Ntpd')


@my_plugin.plugin()
class Ntpd(Plugin):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'stats_dir': config.get('stats_dir', '/var/log/ntpstats')
        }
        self._last_loop_stats = None

    @property
    def last_loop_stats(self):
        loopstats_file = os.path.join(self.config['stats_dir'], 'loopstats')
        if not self._last_loop_stats:
            with open(loopstats_file) as f:
                last_line_items = f.readlines()[-1].split()
            self._last_loop_stats = {
                'timestamp': int(os.stat(loopstats_file).st_mtime),
                'offset': float(last_line_items[2])
            }
        return self._last_loop_stats

    @my_plugin.check()
    def last_sync_delta(self, check_config):
        current_time = int(time.time())
        delta = current_time - self.last_loop_stats['timestamp']
        status = self._value_to_status_less(delta, check_config)
        output = 'Ntp sync {} ago'.format(timedelta(seconds=delta))
        return status, output

    @my_plugin.check()
    def offset(self, check_config):
        status = self._value_to_status_less(
            abs(self.last_loop_stats['offset']), check_config
        )
        output = 'Last time offset: {0:.3f}s'.format(
            self.last_loop_stats['offset']
        )
        return status, output

    @staticmethod
    def config_sample():
        return '''
        # ntpd
        # Enable statistics in /etc/ntp.conf:
        # statsdir /var/log/ntpstats/
        - type: Ntpd
          checks:
            - type: offset
              warn: 0.500
              crit: 2.0
            # Last synchronization data available
            - type: last_sync_delta
              warn: 2800
              crit: 3600
        '''
